use std::rc::Rc;

use yew::html::IntoPropValue;
use yew::prelude::*;
use yew::virtual_dom::{Key, VComp, VNode};

use crate::props::{ContainerBuilder, WidgetBuilder};
use crate::widget::{Column, Container};

use super::NavigationBar;

use pwt_macros::builder;

/// Implements the basic Material Design visual page layout structure.
#[derive(Properties, Clone, PartialEq)]
#[builder]
pub struct Scaffold {
    /// The yew component key.
    #[prop_or_default]
    pub key: Option<Key>,

    /// The top application bar.
    #[prop_or_default]
    pub application_bar: Option<VNode>,

    /// The primary content displayed below the application bar.
    #[prop_or_default]
    pub body: Option<VNode>,

    /// The bottom navigation bar.
    #[builder(IntoPropValue, into_prop_value)]
    #[prop_or_default]
    pub navigation_bar: Option<NavigationBar>,

    /// Favorite action button.
    #[prop_or_default]
    pub favorite_action_button: Option<VNode>,
}

impl Default for Scaffold {
    fn default() -> Self {
        Self::new()
    }
}

impl Scaffold {
    /// Create a new instance.
    pub fn new() -> Self {
        yew::props!(Self {})
    }

    /// Create a scaffold with text title in application bar.
    ///
    /// This is just a cenvenient wrapper for:
    /// ```
    /// # fn test(title: String) {
    /// # use pwt::touch::{Scaffold, ApplicationBar};
    /// Scaffold::new()
    ///   .application_bar(ApplicationBar::new().title(title))
    /// # ;}
    /// ```
    pub fn with_title(title: impl Into<AttrValue>) -> Self {
        let title = title.into();
        Scaffold::new().application_bar(super::ApplicationBar::new().title(title))
    }

    /// Builder style method to set the application bar.
    pub fn application_bar(mut self, app_bar: impl Into<VNode>) -> Self {
        self.application_bar = Some(app_bar.into());
        self
    }

    /// Builder style method to set the body.
    pub fn body(mut self, body: impl Into<VNode>) -> Self {
        self.body = Some(body.into());
        self
    }

    /// Builder style method to set the favorite action button.
    pub fn favorite_action_button(mut self, fav: impl Into<VNode>) -> Self {
        self.favorite_action_button = Some(fav.into());
        self
    }
}

#[doc(hidden)]
pub struct PwtScaffold {}

impl Component for PwtScaffold {
    type Message = ();
    type Properties = Scaffold;

    fn create(_ctx: &Context<Self>) -> Self {
        Self {}
    }

    fn view(&self, ctx: &Context<Self>) -> Html {
        let props = ctx.props();

        let positioned_fab = props.favorite_action_button.clone().map(|fab| {
            Container::new()
                .class("pwt-position-absolute")
                .class("pwt-right-2 pwt-bottom-2")
                .with_child(fab)
        });

        let body = Column::new()
            .class("pwt-position-relative")
            .class("pwt-flex-fill")
            .with_optional_child(props.body.clone())
            .with_optional_child(positioned_fab);

        Column::new()
            .class("pwt-viewport")
            .class("pwt-position-relative")
            .with_optional_child(props.application_bar.clone())
            .with_child(body)
            .with_optional_child(props.navigation_bar.clone())
            .into()
    }
}

impl From<Scaffold> for VNode {
    fn from(val: Scaffold) -> Self {
        let key = val.key.clone();
        let comp = VComp::new::<PwtScaffold>(Rc::new(val), key);
        VNode::from(comp)
    }
}
