use std::marker::PhantomData;
use std::rc::Rc;

use derivative::Derivative;

use yew::virtual_dom::{Key, VComp, VNode};

use crate::prelude::*;

use crate::props::{AsClassesMut, AsCssStylesMut, CssStyles, WidgetBuilder};
use crate::state::DataStore;

use crate::widget::Container;

mod list_render_callback;
pub use list_render_callback::{ListRenderArgs, ListRenderCallback};

mod virtual_scroll;
pub use virtual_scroll::{
    PwtVirtualScroll, VirtualScroll, VirtualScrollInfo, VirtualScrollRenderArgs,
    VirtualScrollRenderCallback,
};

use super::SizeObserver;

use pwt_macros::builder;

/// List with virtual scrolling.
#[derive(Properties, Derivative)]
#[derivative(Clone(bound = ""), PartialEq(bound = ""))]
#[builder]
pub struct List<S: DataStore> {
    // The data collection ([Store] or [TreeStore](crate::state::TreeStore)).
    store: S,
    // List item render function
    renderer: ListRenderCallback<S::Record>,

    /// Yew key property.
    #[prop_or_default]
    pub key: Option<Key>,

    /// CSS class of the container.
    #[prop_or_default]
    pub class: Classes,

    /// CSS styles of the container
    #[prop_or_default]
    pub styles: CssStyles,

    /// Virtual Scroll
    ///
    /// Virtual scroll is enabled by default for tables with more than 30 rows.
    #[prop_or_default]
    #[builder]
    pub virtual_scroll: Option<bool>,

    /// Minimum row height (default 22)
    ///
    /// Sets the minimum height for list rows. This is also used by
    /// the virtual scrolling algorithm to compute the maximal number
    /// of visible rows.
    #[prop_or(22)]
    #[builder]
    pub min_row_height: usize,

    /// Disable horizontal borders.
    #[prop_or_default]
    #[builder]
    pub borderless: bool,

    /// Emphase rows when you mouse over them.
    #[prop_or_default]
    #[builder]
    pub hover: bool,

    /// Use a striped color scheme for rows.
    #[prop_or(true)]
    #[builder]
    pub striped: bool,
}

impl<S: DataStore> List<S> {
    /// Create a new instance.
    ///
    /// The store is either a [Store](crate::state::Store) or [TreeStore](crate::state::TreeStore).
    pub fn new(store: S, renderer: impl Into<ListRenderCallback<S::Record>>) -> Self {
        yew::props!(List<S> { store, renderer: renderer.into() })
    }
    /// Builder style method to set the yew `key` property.
    pub fn key(mut self, key: impl IntoOptionalKey) -> Self {
        self.key = key.into_optional_key();
        self
    }

    /// Builder style method to add a html class.
    pub fn class(mut self, class: impl Into<Classes>) -> Self {
        self.add_class(class);
        self
    }

    /// Method to add a html class.
    pub fn add_class(&mut self, class: impl Into<Classes>) {
        self.class.push(class);
    }
}

impl<S: DataStore> AsClassesMut for List<S> {
    fn as_classes_mut(&mut self) -> &mut yew::Classes {
        &mut self.class
    }
}

impl<S: DataStore> AsCssStylesMut for List<S> {
    fn as_css_styles_mut(&mut self) -> &mut crate::props::CssStyles {
        &mut self.styles
    }
}

impl<S: DataStore> CssBorderBuilder for List<S> {}
impl<S: DataStore> CssPaddingBuilder for List<S> {}
impl<S: DataStore> CssMarginBuilder for List<S> {}
impl<S: DataStore> WidgetStyleBuilder for List<S> {}

pub enum Msg {}

pub struct PwtList<S: DataStore> {
    _phantom_store: PhantomData<S>,
}

impl<S: DataStore> PwtList<S> {
    fn render_list_content(props: &List<S>, start: usize, end: usize) -> Html {
        let mut table = Container::new()
            .attribute("role", "none")
            .class("pwt-list-content")
            // fixme: define list class names
            .class(props.hover.then(|| "table-hover"))
            .class(props.striped.then(|| "table-striped"))
            //.class(props.bordered.then(|| "table-bordered"))
            .class(props.borderless.then(|| "table-borderless"))
            .style("display", "grid")
            .style(
                "row-gap",
                format!("{}px", if props.borderless { 0 } else { 1 }),
            );

        for (filtered_pos, item) in props.store.filtered_data_range(start..end) {
            let record = &*item.record();
            let record_key = props.store.extract_key(record);
            let arg = ListRenderArgs {
                record,
                record_key: &record_key,
                selected: false,
                row_index: filtered_pos,
            };
            let row = props.renderer.apply(&arg);
            table.add_child(row);
        }

        table.into()
    }
}

impl<S: DataStore + 'static> Component for PwtList<S> {
    type Message = Msg;
    type Properties = List<S>;

    fn create(_ctx: &Context<Self>) -> Self {
        Self {
            _phantom_store: PhantomData::<S>,
        }
    }

    fn view(&self, ctx: &Context<Self>) -> Html {
        let props = ctx.props();

        VirtualScroll::new(props.store.filtered_data_len(), {
            let props = props.clone();
            move |args: &VirtualScrollRenderArgs| {
                Self::render_list_content(&props, args.start, args.end)
            }
        })
        .class(props.class.clone())
        .styles(props.styles.clone())
        .into()
    }
}

impl<S: DataStore + 'static> Into<VNode> for List<S> {
    fn into(self) -> VNode {
        let key = self.key.clone();
        let comp = VComp::new::<PwtList<S>>(Rc::new(self), key);
        VNode::from(comp)
    }
}
