use futures::select_biased;
use std::marker::PhantomData;
use std::rc::Rc;
use yew::virtual_dom::{Key, VComp, VNode};

use derivative::Derivative;

use crate::prelude::*;

use crate::props::{
    AsClassesMut, AsCssStylesMut, CssStyles, EventSubscriber, RenderFn, WidgetBuilder,
};
use crate::state::{DataStore, Selection, SelectionObserver};

use crate::widget::Container;

mod list_render_callback;
pub use list_render_callback::{ListRenderArgs, ListRenderCallback};

use super::SizeObserver;

use pwt_macros::{builder, widget};

static VIRTUAL_SCROLL_TRIGGER: usize = 30;

/// List with virtual scrolling.
#[widget(pwt=crate, comp=crate::widget::PwtList<S>, @element)]
#[derive(Properties, Derivative)]
#[derivative(Clone(bound = ""), PartialEq(bound = ""))]
#[builder]
pub struct List<S: DataStore + 'static> {
    // The data collection ([Store] or [TreeStore](crate::state::TreeStore)).
    store: S,
    // List item render function
    renderer: ListRenderCallback<S::Record>,

    /// Yew key property.
    #[prop_or_default]
    pub key: Option<Key>,

    /// Virtual Scroll
    ///
    /// Virtual scroll is enabled by default for tables with more than 30 rows.
    #[prop_or_default]
    #[builder]
    pub virtual_scroll: Option<bool>,

    /// Minimum row height (default 22)
    ///
    /// Sets the minimum height for list rows. This is also used by
    /// the virtual scrolling algorithm to compute the maximal number
    /// of visible rows.
    #[prop_or(22)]
    #[builder]
    pub min_row_height: usize,

    /// Disable horizontal borders.
    #[prop_or_default]
    #[builder]
    pub borderless: bool,

    /// Emphase rows when you mouse over them.
    #[prop_or_default]
    #[builder]
    pub hover: bool,

    /// Use a striped color scheme for rows.
    #[prop_or(true)]
    #[builder]
    pub striped: bool,
}

impl<S: DataStore> List<S> {
    /// Create a new instance.
    ///
    /// The store is either a [Store](crate::state::Store) or [TreeStore](crate::state::TreeStore).
    pub fn new(store: S, renderer: impl Into<ListRenderCallback<S::Record>>) -> Self {
        yew::props!(List<S> { store, renderer: renderer.into() })
    }
    /// Builder style method to set the yew `key` property.
    pub fn key(mut self, key: impl IntoOptionalKey) -> Self {
        self.key = key.into_optional_key();
        self
    }
}

#[derive(Default)]
struct VirtualScrollInfo {
    start: usize,
    end: usize,
    height: f64,
    offset: f64,
    offset_end: f64,
}

impl VirtualScrollInfo {
    fn visible_rows(&self) -> usize {
        self.end.saturating_sub(self.start)
    }

    fn is_visible(&self, row: usize) -> bool {
        (self.start..=self.end).contains(&row)
    }
}

pub enum Msg {
    ScrollTo(i32, i32),
    TableResize(f64, f64),
    ViewportResize(f64, f64, f64),
}

pub struct PwtList<S: DataStore> {
    viewport_height: f64,
    viewport_width: f64,
    viewport_ref: NodeRef,
    viewport_size_observer: Option<SizeObserver>,
    viewport_scrollbar_size: Option<f64>,
    viewport_scroll_top: usize,
    set_viewport_scroll_top: Option<usize>,

    table_ref: NodeRef,
    table_size_observer: Option<SizeObserver>,
    table_height: f64,

    row_height: f64,
    scroll_info: VirtualScrollInfo,

    _phantom_store: PhantomData<S>,
}

impl<S: DataStore> PwtList<S> {
    fn update_scroll_info(&mut self, props: &List<S>) {
        let row_count = props.store.filtered_data_len();

        let virtual_scroll = props
            .virtual_scroll
            .unwrap_or(row_count >= VIRTUAL_SCROLL_TRIGGER);

        let mut start = if virtual_scroll {
            (self.viewport_scroll_top as f64 / self.row_height).floor() as usize
        } else {
            0
        };

        if start > 0 {
            start -= 1;
        }
        if (start & 1) == 1 {
            start -= 1;
        } // make it work with striped rows

        let max_visible_rows =
            (self.viewport_height / props.min_row_height as f64).ceil() as usize + 5;
        let end = if virtual_scroll {
            (start + max_visible_rows).min(row_count)
        } else {
            row_count
        };

        if start > end {
            start = end.saturating_sub(max_visible_rows);
        }

        let offset = (start as f64) * self.row_height;
        let offset_end = offset + self.table_height;

        let height = offset_end + row_count.saturating_sub(end) as f64 * self.row_height;

        self.scroll_info = VirtualScrollInfo {
            start,
            end,
            offset,
            offset_end,
            height,
        };
    }

    fn render_table(&self, props: &List<S>) -> Html {
        let mut table = Container::new()
            .attribute("role", "none")
            .class("pwt-list-content")
            // fixme: define list class names
            .class(props.hover.then(|| "table-hover"))
            .class(props.striped.then(|| "table-striped"))
            //.class(props.bordered.then(|| "table-bordered"))
            //.class(props.borderless.then(|| "table-borderless"))
            .node_ref(self.table_ref.clone())
            .style("position", "relative")
            .style("top", format!("{}px", self.scroll_info.offset));

        for (filtered_pos, item) in props
            .store
            .filtered_data_range(self.scroll_info.start..self.scroll_info.end)
        {
            let record = &*item.record();
            let record_key = props.store.extract_key(record);
            let arg = ListRenderArgs {
                record,
                record_key: &record_key,
                selected: false,
                row_index: filtered_pos,
            };

            if filtered_pos != 0 {
                table.add_child(
                    html! {<div style="height:1px;background-color:var(--pwt-color-border);"/>},
                );
            }

            let row = props.renderer.apply(&arg);
            table.add_child(
                Container::new()
                    .class("pwt-list-item")
                    .style("min-height", "22px")
                    .with_child(row),
            );
        }

        Container::new()
            .height(self.scroll_info.height)
            .attribute("role", "none")
            .with_child(table)
            .into()
    }
}

impl<S: DataStore + 'static> Component for PwtList<S> {
    type Message = Msg;
    type Properties = List<S>;

    fn create(ctx: &Context<Self>) -> Self {
        let props = ctx.props();
        Self {
            _phantom_store: PhantomData::<S>,
            viewport_height: 0.0,
            viewport_width: 0.0,
            viewport_size_observer: None,
            viewport_scrollbar_size: None,
            viewport_ref: NodeRef::default(),
            viewport_scroll_top: 0,
            set_viewport_scroll_top: None,

            table_ref: NodeRef::default(),
            table_size_observer: None,
            table_height: 0.0,
            row_height: props.min_row_height as f64,
            scroll_info: VirtualScrollInfo::default(),
        }
    }

    fn update(&mut self, ctx: &Context<Self>, msg: Self::Message) -> bool {
        let props = ctx.props();
        match msg {
            Msg::ScrollTo(_x, y) => {
                self.viewport_scroll_top = y.max(0) as usize;
                //if let Some(el) = self.header_scroll_ref.cast::<web_sys::Element>() {
                //    el.set_scroll_left(x as i32);
                //}
                self.update_scroll_info(props);
                props.virtual_scroll.unwrap_or(true)
            }
            Msg::ViewportResize(width, height, scrollbar_size) => {
                //log::info!("VIEWPORT RESIZE {} {}", width, height);
                self.viewport_height = height.max(0.0);
                self.viewport_width = width.max(0.0);

                self.viewport_scrollbar_size = if scrollbar_size.abs() < 1.0 {
                    // on certain zoom levels, the scrollbar size calculation is not perfect...
                    None
                } else {
                    Some(scrollbar_size)
                };

                self.update_scroll_info(props);

                true
            }
            Msg::TableResize(_width, height) => {
                //log::info!("Table RESIZE {} {}", _width, height);
                let height = height.max(0.0);
                if self.table_height == height {
                    return false;
                };
                self.table_height = height;
                let visible_rows = self.scroll_info.visible_rows();
                if (height > 0.0) && (visible_rows > 0) {
                    let row_height = height / visible_rows as f64;
                    if row_height > self.row_height {
                        self.row_height = row_height;
                    }
                }
                self.update_scroll_info(props);

                /*
                if self.cursor.is_none() {
                    if let Some(selection) = &props.selection {
                        if let Some(record_key) = selection.selected_key() {
                            self.cursor =
                                self.filtered_record_pos(props, &record_key)
                                    .map(|pos| Cursor {
                                        pos,
                                        record_key: record_key.clone(),
                                    });
                        }
                    }
                    self.scroll_cursor_into_view();
                }
                */

                true
            }
        }
    }

    fn view(&self, ctx: &Context<Self>) -> Html {
        let props = ctx.props();

        let table = self.render_table(props);

        Container::from_widget_props(props.std_props.clone(), Some(props.listeners.clone()))
            .node_ref(self.viewport_ref.clone())
            .class("pwt-list")
            .onscroll(ctx.link().batch_callback(move |event: Event| {
                let target: Option<web_sys::HtmlElement> = event.target_dyn_into();
                target.map(|el| Msg::ScrollTo(el.scroll_left(), el.scroll_top()))
            }))
            .with_child(table)
            .into()
    }

    fn rendered(&mut self, ctx: &Context<Self>, first_render: bool) {
        if first_render {
            if let Some(el) = self.viewport_ref.cast::<web_sys::Element>() {
                let link = ctx.link().clone();
                let size_observer =
                    SizeObserver::new(&el, move |(width, height, client_width, _)| {
                        link.send_message(Msg::ViewportResize(width, height, width - client_width));
                    });
                self.viewport_size_observer = Some(size_observer);
            }
            if let Some(el) = self.table_ref.cast::<web_sys::HtmlElement>() {
                let link = ctx.link().clone();
                let size_observer = SizeObserver::new(&el, move |(width, height)| {
                    link.send_message(Msg::TableResize(width, height));
                });
                self.table_size_observer = Some(size_observer);
            }
        }
        if let Some(top) = self.set_viewport_scroll_top.take() {
            // Note: we delay setting ScrollTop until we rendered the
            // viewport with correct height. Else, set_viewport_scroll_top can
            // fail because the viewport is smaller.
            if let Some(el) = self.viewport_ref.cast::<web_sys::Element>() {
                el.set_scroll_top(top as i32);
            }
        }
    }
}
