use std::rc::Rc;

use derivative::Derivative;
use html::IntoEventCallback;
use yew::virtual_dom::Key;

use super::SizeObserver;
use crate::prelude::*;
use crate::props::{EventSubscriber, WidgetBuilder};
use crate::widget::Container;

use pwt_macros::{builder, widget};

static VIRTUAL_SCROLL_TRIGGER: usize = 30;

/// Virtual scroll render callback arguments.
pub struct VirtualScrollRenderArgs {
    pub start: usize,
    pub end: usize,
}

/// Virtual scroll render callback.
#[derive(Derivative)]
#[derivative(Clone(bound = ""), PartialEq(bound = ""))]
pub struct VirtualScrollRenderCallback(
    #[derivative(PartialEq(compare_with = "Rc::ptr_eq"))]
    Rc<dyn Fn(&VirtualScrollRenderArgs) -> Html>,
);

impl VirtualScrollRenderCallback {
    /// Creates a new instance.
    pub fn new(renderer: impl 'static + Fn(&VirtualScrollRenderArgs) -> Html) -> Self {
        Self(Rc::new(renderer))
    }
    /// Apply the render function
    pub fn apply(&self, args: &VirtualScrollRenderArgs) -> Html {
        (self.0)(args)
    }
}

impl<F: 'static + Fn(&VirtualScrollRenderArgs) -> Html> From<F> for VirtualScrollRenderCallback {
    fn from(f: F) -> Self {
        VirtualScrollRenderCallback::new(f)
    }
}

/// Container with virtual scrolling.
#[widget(pwt=crate, comp=crate::widget::PwtVirtualScroll, @element)]
#[derive(Properties, PartialEq, Clone)]
#[builder]
pub struct VirtualScroll {
    // List item render function
    renderer: VirtualScrollRenderCallback,

    row_count: usize,

    /// Yew key property.
    #[prop_or_default]
    pub key: Option<Key>,

    /// Virtual Scroll
    ///
    /// Virtual scroll is enabled by default for tables with more than 30 rows.
    #[prop_or_default]
    #[builder]
    pub virtual_scroll: Option<bool>,

    /// Minimum row height (default 22)
    ///
    /// Sets the minimum height for list rows. This is also used by
    /// the virtual scrolling algorithm to compute the maximal number
    /// of visible rows.
    #[prop_or(22)]
    #[builder]
    pub min_row_height: usize,

    #[prop_or_default]
    #[builder_cb(IntoEventCallback, into_event_callback, VirtualScrollInfo)]
    pub on_scroll: Option<Callback<VirtualScrollInfo>>,
}

impl VirtualScroll {
    /// Create a new instance.
    ///
    /// The store is either a [Store](crate::state::Store) or [TreeStore](crate::state::TreeStore).
    pub fn new(row_count: usize, renderer: impl Into<VirtualScrollRenderCallback>) -> Self {
        yew::props!(Self {
            row_count,
            renderer: renderer.into()
        })
    }
    /// Builder style method to set the yew `key` property.
    pub fn key(mut self, key: impl IntoOptionalKey) -> Self {
        self.key = key.into_optional_key();
        self
    }
}

#[derive(Default, Clone, PartialEq)]
pub struct VirtualScrollInfo {
    pub start: usize,
    pub end: usize,
    pub height: f64,
    pub offset: f64,
    pub offset_end: f64,
}

impl VirtualScrollInfo {
    pub fn visible_rows(&self) -> usize {
        self.end.saturating_sub(self.start)
    }
    pub fn is_visible(&self, row: usize) -> bool {
        (self.start..=self.end).contains(&row)
    }
}

pub enum Msg {
    ScrollTo(i32, i32),
    TableResize(f64, f64),
    ViewportResize(f64, f64, f64),
}

pub struct PwtVirtualScroll {
    viewport_height: f64,
    viewport_width: f64,
    viewport_ref: NodeRef,
    viewport_size_observer: Option<SizeObserver>,
    viewport_scrollbar_size: Option<f64>,
    viewport_scroll_top: usize,
    set_viewport_scroll_top: Option<usize>,

    table_ref: NodeRef,
    table_size_observer: Option<SizeObserver>,
    table_height: f64,

    row_height: f64,
    scroll_info: VirtualScrollInfo,
}

impl PwtVirtualScroll {
    fn update_scroll_info(&mut self, props: &VirtualScroll) {
        let row_count = props.row_count;

        let virtual_scroll = props
            .virtual_scroll
            .unwrap_or(row_count >= VIRTUAL_SCROLL_TRIGGER);

        let mut start = if virtual_scroll {
            (self.viewport_scroll_top as f64 / self.row_height).floor() as usize
        } else {
            0
        };

        if start > 0 {
            start -= 1;
        }
        if (start & 1) == 1 {
            start -= 1;
        } // make it work with striped rows

        let max_visible_rows =
            (self.viewport_height / props.min_row_height as f64).ceil() as usize + 5;
        let end = if virtual_scroll {
            (start + max_visible_rows).min(row_count)
        } else {
            row_count
        };

        if start > end {
            start = end.saturating_sub(max_visible_rows);
        }

        let offset = (start as f64) * self.row_height;
        let offset_end = offset + self.table_height;

        let height = offset_end + row_count.saturating_sub(end) as f64 * self.row_height;

        self.scroll_info = VirtualScrollInfo {
            start,
            end,
            offset,
            offset_end,
            height,
        };

        if let Some(on_scroll) = &props.on_scroll {
            on_scroll.emit(self.scroll_info.clone());
        }
    }

    fn render_content(&self, props: &VirtualScroll) -> Html {
        let mut scrolled_content = Container::new()
            .attribute("role", "none")
            .class("pwt-virtual-scroll-content")
            .node_ref(self.table_ref.clone())
            .style("position", "relative")
            .style("top", format!("{}px", self.scroll_info.offset));

        let arg = VirtualScrollRenderArgs {
            start: self.scroll_info.start,
            end: self.scroll_info.end,
        };

        let content = props.renderer.apply(&arg);
        scrolled_content.add_child(content);

        Container::new()
            .height(self.scroll_info.height)
            .attribute("role", "none")
            .with_child(scrolled_content)
            .into()
    }
}

impl Component for PwtVirtualScroll {
    type Message = Msg;
    type Properties = VirtualScroll;

    fn create(ctx: &Context<Self>) -> Self {
        let props = ctx.props();
        Self {
            viewport_height: 0.0,
            viewport_width: 0.0,
            viewport_size_observer: None,
            viewport_scrollbar_size: None,
            viewport_ref: NodeRef::default(),
            viewport_scroll_top: 0,
            set_viewport_scroll_top: None,

            table_ref: NodeRef::default(),
            table_size_observer: None,
            table_height: 0.0,
            row_height: props.min_row_height as f64,
            scroll_info: VirtualScrollInfo::default(),
        }
    }

    fn changed(&mut self, ctx: &Context<Self>, _old_props: &Self::Properties) -> bool {
        let props = ctx.props();
        self.update_scroll_info(props);
        true
    }

    fn update(&mut self, ctx: &Context<Self>, msg: Self::Message) -> bool {
        let props = ctx.props();
        match msg {
            Msg::ScrollTo(_x, y) => {
                self.viewport_scroll_top = y.max(0) as usize;
                //if let Some(el) = self.header_scroll_ref.cast::<web_sys::Element>() {
                //    el.set_scroll_left(x as i32);
                //}
                self.update_scroll_info(props);
                props.virtual_scroll.unwrap_or(true)
            }
            Msg::ViewportResize(width, height, scrollbar_size) => {
                //log::info!("VIEWPORT RESIZE {} {}", width, height);
                self.viewport_height = height.max(0.0);
                self.viewport_width = width.max(0.0);

                self.viewport_scrollbar_size = if scrollbar_size.abs() < 1.0 {
                    // on certain zoom levels, the scrollbar size calculation is not perfect...
                    None
                } else {
                    Some(scrollbar_size)
                };

                self.update_scroll_info(props);

                true
            }
            Msg::TableResize(_width, height) => {
                //log::info!("Table RESIZE {} {}", _width, height);
                let height = height.max(0.0);
                if self.table_height == height {
                    return false;
                };
                self.table_height = height;
                let visible_rows = self.scroll_info.visible_rows();
                if (height > 0.0) && (visible_rows > 0) {
                    let row_height = height / visible_rows as f64;
                    if row_height > self.row_height {
                        self.row_height = row_height;
                    }
                }
                self.update_scroll_info(props);

                /*
                if self.cursor.is_none() {
                    if let Some(selection) = &props.selection {
                        if let Some(record_key) = selection.selected_key() {
                            self.cursor =
                                self.filtered_record_pos(props, &record_key)
                                    .map(|pos| Cursor {
                                        pos,
                                        record_key: record_key.clone(),
                                    });
                        }
                    }
                    self.scroll_cursor_into_view();
                }
                */

                true
            }
        }
    }

    fn view(&self, ctx: &Context<Self>) -> Html {
        let props = ctx.props();

        let table = self.render_content(props);

        Container::from_widget_props(props.std_props.clone(), Some(props.listeners.clone()))
            .node_ref(self.viewport_ref.clone())
            .class("pwt-virtual-scroll")
            .onscroll(ctx.link().batch_callback(move |event: Event| {
                let target: Option<web_sys::HtmlElement> = event.target_dyn_into();
                target.map(|el| Msg::ScrollTo(el.scroll_left(), el.scroll_top()))
            }))
            .with_child(table)
            .into()
    }

    fn rendered(&mut self, ctx: &Context<Self>, first_render: bool) {
        if first_render {
            if let Some(el) = self.viewport_ref.cast::<web_sys::Element>() {
                let link = ctx.link().clone();
                let size_observer =
                    SizeObserver::new(&el, move |(width, height, client_width, _)| {
                        link.send_message(Msg::ViewportResize(width, height, width - client_width));
                    });
                self.viewport_size_observer = Some(size_observer);
            }
            if let Some(el) = self.table_ref.cast::<web_sys::HtmlElement>() {
                let link = ctx.link().clone();
                let size_observer = SizeObserver::new(&el, move |(width, height)| {
                    link.send_message(Msg::TableResize(width, height));
                });
                self.table_size_observer = Some(size_observer);
            }
        }
        if let Some(top) = self.set_viewport_scroll_top.take() {
            // Note: we delay setting ScrollTop until we rendered the
            // viewport with correct height. Else, set_viewport_scroll_top can
            // fail because the viewport is smaller.
            if let Some(el) = self.viewport_ref.cast::<web_sys::Element>() {
                el.set_scroll_top(top as i32);
            }
        }
    }
}
